@@ -4,7 +4,7 @@ This module provides SQL query parsing functions
 """
 import logging
 import re
-from typing import Dict, List, Optional, Tuple, Union
+from typing import Dict, List, Optional, Set, Tuple, Union
 
 import sqlparse
 from sqlparse.sql import Token
@@ -14,7 +14,7 @@ from sql_metadata.generalizator import Generalizator
 from sql_metadata.keywords_lists import (
     COLUMNS_SECTIONS,
     KEYWORDS_BEFORE_COLUMNS,
-    QueryType,
+    TokenType,
     RELEVANT_KEYWORDS,
     SUBQUERY_PRECEDING_KEYWORDS,
     SUPPORTED_QUERY_TYPES,
@@ -43,7 +43,7 @@ class Parser:  # pylint: disable=R0902
         self._columns_dict = None
         self._columns_aliases_names = None
         self._columns_aliases = None
-        self._columns_with_tables_aliases = dict()
+        self._columns_with_tables_aliases = {}
         self._columns_aliases_dict = None
 
         self._tables = None
@@ -54,8 +54,8 @@ class Parser:  # pylint: disable=R0902
         self._with_queries_columns = None
         self._subqueries = None
         self._subqueries_names = None
-        self._subqueries_parsers = dict()
-        self._with_parsers = dict()
+        self._subqueries_parsers = {}
+        self._with_parsers = {}
 
         self._limit_and_offset = None
 
@@ -65,14 +65,16 @@ class Parser:  # pylint: disable=R0902
         self._subquery_level = 0
         self._nested_level = 0
         self._parenthesis_level = 0
-        self._open_parentheses = []
+        self._open_parentheses: List[SQLToken] = []
         self._aliases_to_check = None
         self._is_in_nested_function = False
         self._is_in_with_block = False
-        self._with_columns_candidates = dict()
-        self._column_aliases_max_subquery_level = dict()
+        self._with_columns_candidates = {}
+        self._column_aliases_max_subquery_level = {}
 
         self.sqlparse_tokens = None
+        self.non_empty_tokens = None
+        self.tokens_length = None
 
     @property
     def query(self) -> str:
@@ -101,10 +103,19 @@ class Parser:  # pylint: disable=R0902
         if not tokens:
             raise ValueError("Empty queries are not supported!")
 
-        if tokens[0].normalized in ["CREATE", "ALTER"]:
-            switch = tokens[0].normalized + tokens[1].normalized
+        index = (
+            0
+            if not tokens[0].is_left_parenthesis
+            else tokens[0]
+            .find_nearest_token(
+                value=False, value_attribute="is_left_parenthesis", direction="right"
+            )
+            .position
+        )
+        if tokens[index].normalized in ["CREATE", "ALTER"]:
+            switch = tokens[index].normalized + tokens[index + 1].normalized
         else:
-            switch = tokens[0].normalized
+            switch = tokens[index].normalized
         self._query_type = SUPPORTED_QUERY_TYPES.get(switch, "UNSUPPORTED")
         if self._query_type == "UNSUPPORTED":
             self._logger.error("Not supported query type: %s", self._raw_query)
@@ -124,34 +135,38 @@ class Parser:  # pylint: disable=R0902
         # handle empty queries (#12)
         if not parsed:
             return tokens
-
-        self.sqlparse_tokens = parsed[0].tokens
-        sqlparse_tokens = self._flatten_sqlparse()
-        non_empty_tokens = [
-            token
-            for token in sqlparse_tokens
-            if token.ttype is not Whitespace and token.ttype.parent is not Whitespace
-        ]
+        self._get_sqlparse_tokens(parsed)
         last_keyword = None
-        for index, tok in enumerate(non_empty_tokens):
+        combine_flag = False
+        for index, tok in enumerate(self.non_empty_tokens):
+            # combine dot separated identifiers
+            if self._is_token_part_of_complex_identifier(token=tok, index=index):
+                combine_flag = True
+                continue
             token = SQLToken(
                 tok=tok,
                 index=index,
                 subquery_level=self._subquery_level,
                 last_keyword=last_keyword,
             )
-            if index > 0:
-                # create links between consecutive tokens
-                token.previous_token = tokens[index - 1]
-                tokens[index - 1].next_token = token
+            if combine_flag:
+                self._combine_qualified_names(index=index, token=token)
+                combine_flag = False
+
+            previous_token = tokens[-1] if index > 0 else EmptyToken
+            token.previous_token = previous_token
+            previous_token.next_token = token if index > 0 else None
 
             if token.is_left_parenthesis:
+                token.token_type = TokenType.PARENTHESIS
                 self._determine_opening_parenthesis_type(token=token)
             elif token.is_right_parenthesis:
+                token.token_type = TokenType.PARENTHESIS
                 self._determine_closing_parenthesis_type(token=token)
 
-            if tok.is_keyword and "".join(tok.normalized.split()) in RELEVANT_KEYWORDS:
-                last_keyword = tok.normalized
+            last_keyword = self._determine_last_relevant_keyword(
+                token=token, last_keyword=last_keyword
+            )
             token.is_in_nested_function = self._is_in_nested_function
             token.parenthesis_level = self._parenthesis_level
             tokens.append(token)
@@ -172,101 +187,40 @@ class Parser:  # pylint: disable=R0902
         if self._columns is not None:
             return self._columns
         columns = UniqueList()
-        tables_aliases = self.tables_aliases
-        subqueries_names = self.subqueries_names
-        with_names = self.with_names
 
-        for token in self.tokens:
-            # handle CREATE TABLE queries (#35)
-            if token.is_name and self.query_type == QueryType.CREATE:
-                # previous token is either ( or , -> indicates the column name
-                if (
-                    token.is_in_parenthesis
-                    and token.previous_token.is_punctuation
-                    and token.last_keyword_normalized == "TABLE"
+        for token in self._not_parsed_tokens:
+            if token.is_name or token.is_keyword_column_name:
+                if token.is_column_definition_inside_create_table(
+                    query_type=self.query_type
                 ):
+                    token.token_type = TokenType.COLUMN
                     columns.append(token.value)
-                    continue
-
-                # we're in CREATE TABLE query with the columns
-                # ignore any annotations outside the parenthesis with the list of columns
-                # e.g. ) CHARACTER SET utf8;
-                if (
-                    not token.is_in_parenthesis
-                    and token.find_nearest_token("SELECT", value_attribute="normalized")
-                    is EmptyToken
-                ):
-                    continue
-
-            if (
-                token.is_name and not token.next_token.is_dot
-            ) or token.is_keyword_column_name:
-                # analyze the name tokens, column names and where condition values
-                if (
-                    token.last_keyword_normalized in KEYWORDS_BEFORE_COLUMNS
-                    and token.previous_token.normalized not in ["AS", ")"]
-                    and not token.is_alias_without_as
-                    and (
-                        token.left_expanded not in self.columns_aliases_names
-                        or token.token_is_alias_of_self_not_from_subquery(
-                            aliases_levels=self._column_aliases_max_subquery_level
-                        )
-                    )
-                ):
-
-                    if (
-                        not (
-                            # aliases of sub-queries i.e.: SELECT from (...) <alias>
-                            token.previous_token.is_right_parenthesis
-                            and token.value in subqueries_names
-                        )
-                        and not (
-                            # names of the with queries <name> as (subquery)
-                            token.next_token.normalized == "AS"
-                            and token.value in with_names
-                        )
-                        # custom functions - they are followed by the parenthesis
-                        # e.g. custom_func(...
-                        and not token.next_token.is_left_parenthesis
-                    ):
-                        column = token.table_prefixed_column(tables_aliases)
-                        if self._is_with_query_already_resolved(column):
-                            self._add_to_columns_aliases_subsection(
-                                token=token, left_expand=False
-                            )
-                            continue
-                        column = self._resolve_sub_queries(column)
-                        self._add_to_columns_with_tables(token, column)
-                        self._add_to_columns_subsection(
-                            keyword=token.last_keyword_normalized, column=column
-                        )
-                        columns.extend(column)
-
                 elif (
-                    token.last_keyword_normalized == "INTO"
-                    and token.previous_token.is_punctuation
+                    token.is_potential_column_name
+                    and token.is_not_an_alias_or_is_self_alias_outside_of_subquery(
+                        columns_aliases_names=self.columns_aliases_names,
+                        max_subquery_level=self._column_aliases_max_subquery_level,
+                    )
+                    and not token.is_sub_query_name_or_with_name_or_function_name(
+                        sub_queries_names=self.subqueries_names,
+                        with_names=self.with_names,
+                    )
+                    and not token.is_table_definition_suffix_in_non_select_create_table(
+                        query_type=self.query_type
+                    )
+                    and not token.is_conversion_specifier
                 ):
-                    # INSERT INTO `foo` (col1, `col2`) VALUES (..)
+                    self._handle_column_save(token=token, columns=columns)
+
+                elif token.is_column_name_inside_insert_clause:
                     column = str(token.value).strip("`")
                     self._add_to_columns_subsection(
                         keyword=token.last_keyword_normalized, column=column
                     )
+                    token.token_type = TokenType.COLUMN
                     columns.append(column)
-                elif token.left_expanded in self.columns_aliases_names:
-                    self._add_to_columns_aliases_subsection(token=token)
-            elif (
-                token.is_wildcard
-                and token.last_keyword_normalized == "SELECT"
-                and not token.previous_token.is_left_parenthesis
-            ):
-                # handle * wildcard in select part, but ignore count(*)
-                column = token.table_prefixed_column(tables_aliases)
-                column = self._resolve_sub_queries(column)
-                self._add_to_columns_with_tables(token, column)
-                self._add_to_columns_subsection(
-                    keyword=token.last_keyword_normalized, column=column
-                )
-                columns.extend(column)
+            elif token.is_a_wildcard_in_select_statement:
+                self._handle_column_save(token=token, columns=columns)
 
         self._columns = columns
         return self._columns
@@ -303,7 +257,7 @@ class Parser:  # pylint: disable=R0902
         """
         if self._columns_aliases is not None:
             return self._columns_aliases
-        column_aliases = dict()
+        column_aliases = {}
         _ = self.columns
         self._aliases_to_check = (
             list(self._columns_with_tables_aliases.keys())
@@ -311,55 +265,33 @@ class Parser:  # pylint: disable=R0902
             + ["*"]
         )
         for token in self.tokens:
-            if (
-                token.value in self.columns_aliases_names
-                and token.value not in column_aliases
-                and not token.previous_token.is_nested_function_start
-                and token.is_alias_definition
+            if token.is_potential_column_alias(
+                column_aliases=column_aliases,
+                columns_aliases_names=self.columns_aliases_names,
             ):
-                if token.previous_token.normalized == "AS":
-                    token_check = token.get_nth_previous(2)
-                else:
-                    token_check = token.previous_token
+                token_check = (
+                    token.previous_token
+                    if not token.previous_token.normalized == "AS"
+                    else token.get_nth_previous(2)
+                )
                 if token_check.is_column_definition_end:
-                    # nested subquery like select a, (select a as b from x) as column
+                    alias_of = self._resolve_subquery_alias(token=token)
+                elif token_check.is_partition_clause_end:
                     start_token = token.find_nearest_token(
-                        True, value_attribute="is_column_definition_start"
+                        True, value_attribute="is_partition_clause_start"
+                    )
+                    alias_of = self._find_all_columns_between_tokens(
+                        start_token=start_token, end_token=token
                     )
-                    if start_token.next_token.normalized == "SELECT":
-                        # we have a subquery
-                        alias_token = start_token.next_token.find_nearest_token(
-                            self._aliases_to_check,
-                            direction="right",
-                            value_attribute="left_expanded",
-                        )
-                        alias_of = self._resolve_alias_to_column(alias_token)
-                    else:
-                        # chain of functions or redundant parenthesis
-                        alias_of = self._find_all_columns_between_tokens(
-                            start_token=start_token, end_token=token
-                        )
                 elif token.is_in_with_columns:
                     # columns definition is to the right in subquery
                     # we are in: with with_name (<aliases>) as (subquery)
                     alias_of = self._find_column_for_with_column_alias(token)
                 else:
-                    # it can be one function or a chain of functions
-                    # like: sum(a) + sum(b) as alias
-                    # or operation on columns like: col1 + col2 as alias
-                    start_token = token.find_nearest_token(
-                        [",", "SELECT"], value_attribute="normalized"
-                    )
-                    while start_token.is_in_nested_function:
-                        start_token = start_token.find_nearest_token(
-                            [",", "SELECT"], value_attribute="normalized"
-                        )
-                    alias_of = self._find_all_columns_between_tokens(
-                        start_token=start_token, end_token=token
-                    )
+                    alias_of = self._resolve_function_alias(token=token)
                 if token.value != alias_of:
                     # skip aliases of self, like sum(column) as column
-                    column_aliases[token.left_expanded] = alias_of
+                    column_aliases[token.value] = alias_of
 
         self._columns_aliases = column_aliases
         return self._columns_aliases
@@ -387,26 +319,16 @@ class Parser:  # pylint: disable=R0902
         column_aliases_names = UniqueList()
         with_names = self.with_names
         subqueries_names = self.subqueries_names
-        for token in self.tokens:
-            if (
-                token.is_name
-                or (token.is_keyword and token.previous_token.normalized == "AS")
-            ) and not token.next_token.is_dot:
-                if (
-                    token.last_keyword_normalized in KEYWORDS_BEFORE_COLUMNS
-                    and token.normalized not in ["DIV"]
-                    and token.is_alias_definition
-                    or token.is_in_with_columns
-                ) and token.value not in with_names + subqueries_names:
-                    alias = token.left_expanded
-                    column_aliases_names.append(alias)
-                    current_level = self._column_aliases_max_subquery_level.setdefault(
-                        alias, 0
-                    )
-                    if token.subquery_level > current_level:
-                        self._column_aliases_max_subquery_level[
-                            alias
-                        ] = token.subquery_level
+        for token in self._not_parsed_tokens:
+            if token.is_potential_alias:
+                if token.value in column_aliases_names:
+                    self._handle_column_alias_subquery_level_update(token=token)
+                elif (
+                    token.is_a_valid_alias
+                    and token.value not in with_names + subqueries_names
+                ):
+                    column_aliases_names.append(token.value)
+                    self._handle_column_alias_subquery_level_update(token=token)
 
         self._columns_aliases_names = column_aliases_names
         return self._columns_aliases_names
@@ -421,51 +343,22 @@ class Parser:  # pylint: disable=R0902
         tables = UniqueList()
         with_names = self.with_names
 
-        for token in self.tokens:
-            if (
-                (token.is_name or token.is_keyword)
-                and token.last_keyword_normalized in TABLE_ADJUSTMENT_KEYWORDS
-                and token.previous_token.normalized not in ["AS", "WITH"]
-                and token.normalized not in ["AS", "SELECT"]
-            ):
-                # handle CREATE TABLE queries (#35)
-                # skip keyword that are withing parenthesis-wrapped list of column
+        for token in self._not_parsed_tokens:
+            if token.is_potential_table_name:
                 if (
-                    self.query_type == QueryType.CREATE
-                    and token.is_in_parenthesis
-                    and token.is_create_table_columns_definition
+                    token.is_alias_of_table_or_alias_of_subquery
+                    or token.is_with_statement_nested_in_subquery
+                    or token.is_constraint_definition_inside_create_table_clause(
+                        query_type=self.query_type
+                    )
+                    or token.is_columns_alias_of_with_query_or_column_in_insert_query(
+                        with_names=with_names
+                    )
                 ):
                     continue
-                if (
-                    token.normalized == "WITH"
-                    and token.previous_token.is_left_parenthesis
-                    and token.get_nth_previous(2).normalized == "FROM"
-                ):
-                    continue
-
-                if token.next_token.is_dot:
-                    pass  # part of the qualified name
-                elif token.is_in_parenthesis and (
-                    token.find_nearest_token("(").previous_token.value in with_names
-                    or token.last_keyword_normalized == "INTO"
-                ):
-                    # we are in <columns> of INSERT INTO <TABLE> (<columns>)
-                    # or columns of with statement: with (<columns>) as ...
-                    pass
-                elif token.previous_token.is_dot:
-                    tables.append(token.left_expanded)  # full qualified name
-                elif (
-                    token.previous_token.normalized != token.last_keyword_normalized
-                    and not token.previous_token.is_punctuation
-                ) or token.previous_token.is_right_parenthesis:
-                    # it's not a list of tables, e.g. SELECT * FROM foo, bar
-                    # hence, it can be the case of alias without AS,
-                    # e.g. SELECT * FROM foo bar
-                    # or an alias of subquery (SELECT * FROM foo) bar
-                    pass
-                else:
-                    table_name = str(token.value.strip("`"))
-                    tables.append(table_name)
+                table_name = str(token.value.strip("`"))
+                token.token_type = TokenType.TABLE
+                tables.append(table_name)
 
         self._tables = tables - with_names
         return self._tables
@@ -480,7 +373,7 @@ class Parser:  # pylint: disable=R0902
         limit = None
         offset = None
 
-        for token in self.tokens:
+        for token in self._not_parsed_tokens:
             if token.is_integer:
                 if token.last_keyword_normalized == "LIMIT" and not limit:
                     # LIMIT <limit>
@@ -509,24 +402,24 @@ class Parser:  # pylint: disable=R0902
         """
         if self._table_aliases is not None:
             return self._table_aliases
-        aliases = dict()
+        aliases = {}
         tables = self.tables
 
-        for token in self.tokens:
+        for token in self._not_parsed_tokens:
             if (
                 token.last_keyword_normalized in TABLE_ADJUSTMENT_KEYWORDS
                 and token.is_name
                 and token.next_token.normalized != "AS"
-                and not token.next_token.is_dot
             ):
                 if token.previous_token.normalized == "AS":
                     # potential <DB.<SCHEMA>.<TABLE> as <ALIAS>
-                    potential_table_name = token.get_nth_previous(2).left_expanded
+                    potential_table_name = token.get_nth_previous(2).value
                 else:
                     # potential <DB.<SCHEMA>.<TABLE> <ALIAS>
-                    potential_table_name = token.previous_token.left_expanded
+                    potential_table_name = token.previous_token.value
 
                 if potential_table_name in tables:
+                    token.token_type = TokenType.TABLE_ALIAS
                     aliases[token.value] = potential_table_name
 
         self._table_aliases = aliases
@@ -545,25 +438,12 @@ class Parser:  # pylint: disable=R0902
         if self._with_names is not None:
             return self._with_names
         with_names = UniqueList()
-        for token in self.tokens:
+        for token in self._not_parsed_tokens:
             if token.previous_token.normalized == "WITH":
                 self._is_in_with_block = True
                 while self._is_in_with_block and token.next_token:
-                    # name is first
                     if token.next_token.normalized == "AS":
-                        if token.is_right_parenthesis:
-                            # inside columns of with statement
-                            # like: with (col1, col2) as (subquery)
-                            token.is_with_columns_end = True
-                            token.is_nested_function_end = False
-                            start_token = token.find_nearest_token("(")
-                            start_token.is_with_columns_start = True
-                            start_token.is_nested_function_start = False
-                            prev_token = start_token.previous_token
-                            with_names.append(prev_token.left_expanded)
-                        else:
-                            with_names.append(token.left_expanded)
-                        # move to next with query end
+                        self._handle_with_name_save(token=token, with_names=with_names)
                         while token.next_token and not token.is_with_query_end:
                             token = token.next_token
                         if token.next_token.normalized in WITH_ENDING_KEYWORDS:
@@ -587,11 +467,11 @@ class Parser:  # pylint: disable=R0902
         """
         if self._with_queries is not None:
             return self._with_queries
-        with_queries = dict()
-        with_queries_columns = dict()
+        with_queries = {}
+        with_queries_columns = {}
         for name in self.with_names:
             token = self.tokens[0].find_nearest_token(
-                name, value_attribute="left_expanded", direction="right"
+                name, value_attribute="value", direction="right"
             )
             if token.next_token.is_with_columns_start:
                 with_queries_columns[name] = True
@@ -621,7 +501,7 @@ class Parser:  # pylint: disable=R0902
         """
         if self._subqueries is not None:
             return self._subqueries
-        subqueries = dict()
+        subqueries = {}
         token = self.tokens[0]
         while token.next_token:
             if token.previous_token.is_subquery_start:
@@ -665,6 +545,7 @@ class Parser:  # pylint: disable=R0902
                 token.previous_token.normalized == "AS"
                 and token.get_nth_previous(2).is_subquery_end
             ):
+                token.token_type = TokenType.SUB_QUERY_NAME
                 subqueries_names.append(str(token))
 
         self._subqueries_names = subqueries_names
@@ -678,7 +559,7 @@ class Parser:  # pylint: disable=R0902
         if self._values:
             return self._values
         values = []
-        for token in self.tokens:
+        for token in self._not_parsed_tokens:
             if (
                 token.last_keyword_normalized == "VALUES"
                 and token.is_in_parenthesis
@@ -734,12 +615,93 @@ class Parser:  # pylint: disable=R0902
         """
         return Generalizator(self._raw_query).generalize
 
+    @property
+    def _not_parsed_tokens(self):
+        """
+        Returns only tokens that have no type assigned yet
+        """
+        return [x for x in self.tokens if x.token_type is None]
+
+    def _handle_column_save(self, token: SQLToken, columns: List[str]):
+        column = token.table_prefixed_column(self.tables_aliases)
+        if self._is_with_query_already_resolved(column):
+            self._add_to_columns_aliases_subsection(token=token, left_expand=False)
+            token.token_type = TokenType.COLUMN_ALIAS
+            return
+        column = self._resolve_sub_queries(column)
+        self._add_to_columns_with_tables(token, column)
+        self._add_to_columns_subsection(
+            keyword=token.last_keyword_normalized, column=column
+        )
+        token.token_type = TokenType.COLUMN
+        columns.extend(column)
+
+    @staticmethod
+    def _handle_with_name_save(token: SQLToken, with_names: List[str]) -> None:
+        if token.is_right_parenthesis:
+            # inside columns of with statement
+            # like: with (col1, col2) as (subquery)
+            token.is_with_columns_end = True
+            token.is_nested_function_end = False
+            start_token = token.find_nearest_token("(")
+            start_token.is_with_columns_start = True
+            start_token.is_nested_function_start = False
+            prev_token = start_token.previous_token
+            prev_token.token_type = TokenType.WITH_NAME
+            with_names.append(prev_token.value)
+        else:
+            token.token_type = TokenType.WITH_NAME
+            with_names.append(token.value)
+
+    def _handle_column_alias_subquery_level_update(self, token: SQLToken) -> None:
+        token.token_type = TokenType.COLUMN_ALIAS
+        self._add_to_columns_aliases_subsection(token=token)
+        current_level = self._column_aliases_max_subquery_level.setdefault(
+            token.value, 0
+        )
+        if token.subquery_level > current_level:
+            self._column_aliases_max_subquery_level[token.value] = token.subquery_level
+
+    def _resolve_subquery_alias(self, token: SQLToken) -> Union[str, List[str]]:
+        # nested subquery like select a, (select a as b from x) as column
+        start_token = token.find_nearest_token(
+            True, value_attribute="is_column_definition_start"
+        )
+        if start_token.next_token.normalized == "SELECT":
+            # we have a subquery
+            alias_token = start_token.next_token.find_nearest_token(
+                self._aliases_to_check,
+                direction="right",
+                value_attribute="value",
+            )
+            return self._resolve_alias_to_column(alias_token)
+
+        # chain of functions or redundant parenthesis
+        return self._find_all_columns_between_tokens(
+            start_token=start_token, end_token=token
+        )
+
+    def _resolve_function_alias(self, token: SQLToken) -> Union[str, List[str]]:
+        # it can be one function or a chain of functions
+        # like: sum(a) + sum(b) as alias
+        # or operation on columns like: col1 + col2 as alias
+        start_token = token.find_nearest_token(
+            [",", "SELECT"], value_attribute="normalized"
+        )
+        while start_token.is_in_nested_function:
+            start_token = start_token.find_nearest_token(
+                [",", "SELECT"], value_attribute="normalized"
+            )
+        return self._find_all_columns_between_tokens(
+            start_token=start_token, end_token=token
+        )
+
     def _add_to_columns_subsection(self, keyword: str, column: Union[str, List[str]]):
         """
         Add columns to the section in which it appears in query
         """
         section = COLUMNS_SECTIONS[keyword]
-        self._columns_dict = self._columns_dict or dict()
+        self._columns_dict = self._columns_dict or {}
         current_section = self._columns_dict.setdefault(section, UniqueList())
         if isinstance(column, str):
             current_section.append(column)
@@ -753,14 +715,14 @@ class Parser:  # pylint: disable=R0902
         Add alias to the section in which it appears in query
         """
         keyword = token.last_keyword_normalized
-        alias = token.left_expanded if left_expand else token.value
+        alias = token.value if left_expand else token.value.split(".")[-1]
         if (
             token.last_keyword_normalized in ["FROM", "WITH"]
             and token.find_nearest_token("(").is_with_columns_start
         ):
             keyword = "SELECT"
         section = COLUMNS_SECTIONS[keyword]
-        self._columns_aliases_dict = self._columns_aliases_dict or dict()
+        self._columns_aliases_dict = self._columns_aliases_dict or {}
         self._columns_aliases_dict.setdefault(section, UniqueList()).append(alias)
 
     def _add_to_columns_with_tables(
@@ -768,28 +730,32 @@ class Parser:  # pylint: disable=R0902
     ) -> None:
         if isinstance(column, list) and len(column) == 1:
             column = column[0]
-        self._columns_with_tables_aliases[token.left_expanded] = column
+        self._columns_with_tables_aliases[token.value] = column
 
-    def _resolve_column_alias(self, alias: Union[str, List[str]]) -> Union[str, List]:
+    def _resolve_column_alias(
+        self, alias: Union[str, List[str]], visited: Set = None
+    ) -> Union[str, List]:
         """
         Returns a column name for a given alias
         """
+        visited = visited or set()
         if isinstance(alias, list):
-            return [self._resolve_column_alias(x) for x in alias]
-        while alias in self.columns_aliases:
+            return [self._resolve_column_alias(x, visited) for x in alias]
+        while alias in self.columns_aliases and alias not in visited:
+            visited.add(alias)
             alias = self.columns_aliases[alias]
             if isinstance(alias, list):
-                return self._resolve_column_alias(alias)
+                return self._resolve_column_alias(alias, visited)
         return alias
 
     def _resolve_alias_to_column(self, alias_token: SQLToken) -> str:
         """
         Resolves aliases of tables to already resolved columns
         """
-        if alias_token.left_expanded in self._columns_with_tables_aliases:
-            alias_of = self._columns_with_tables_aliases[alias_token.left_expanded]
+        if alias_token.value in self._columns_with_tables_aliases:
+            alias_of = self._columns_with_tables_aliases[alias_token.value]
         else:
-            alias_of = alias_token.left_expanded
+            alias_of = alias_token.value
         return alias_of
 
     def _resolve_sub_queries(self, column: str) -> List[str]:
@@ -862,7 +828,7 @@ class Parser:  # pylint: disable=R0902
         parts = col_alias.split(".")
         if len(parts) != 2 or parts[0] not in self.with_names:
             return False
-        if self._with_queries_columns[parts[0]]:
+        if self._with_queries_columns.get(parts[0]):
             return True
         return False
 
@@ -880,11 +846,13 @@ class Parser:  # pylint: disable=R0902
             token.is_column_definition_start = True
         elif token.previous_token.normalized == "AS":
             token.is_with_query_start = True
-        elif token.last_keyword_normalized == "TABLE" and (
-            token.get_nth_previous(2).normalized == "TABLE"
-            or token.get_nth_previous(4).normalized == "TABLE"
+        elif (
+            token.last_keyword_normalized == "TABLE"
+            and token.find_nearest_token("(") is EmptyToken
         ):
             token.is_create_table_columns_declaration_start = True
+        elif token.previous_token.normalized == "OVER":
+            token.is_partition_clause_start = True
         else:
             # nested function
             token.is_nested_function_start = True
@@ -907,6 +875,8 @@ class Parser:  # pylint: disable=R0902
             token.is_with_query_end = True
         elif last_open_parenthesis.is_create_table_columns_declaration_start:
             token.is_create_table_columns_declaration_end = True
+        elif last_open_parenthesis.is_partition_clause_start:
+            token.is_partition_clause_end = True
         else:
             token.is_nested_function_end = True
             self._nested_level -= 1
@@ -941,7 +911,7 @@ class Parser:  # pylint: disable=R0902
         loop_token = start_token
         aliases = UniqueList()
         while loop_token.next_token != end_token:
-            if loop_token.next_token.left_expanded in self._aliases_to_check:
+            if loop_token.next_token.value in self._aliases_to_check:
                 alias_token = loop_token.next_token
                 if (
                     alias_token.normalized != "*"
@@ -979,6 +949,65 @@ class Parser:  # pylint: disable=R0902
 
         return query
 
+    def _determine_last_relevant_keyword(self, token: SQLToken, last_keyword: str):
+        if token.is_keyword and "".join(token.normalized.split()) in RELEVANT_KEYWORDS:
+            if not (
+                token.normalized == "FROM"
+                and token.get_nth_previous(3).normalized == "EXTRACT"
+            ) and not (
+                token.normalized == "ORDERBY"
+                and len(self._open_parentheses) > 0
+                and self._open_parentheses[-1].is_partition_clause_start
+            ):
+                last_keyword = token.normalized
+        return last_keyword
+
+    def _is_token_part_of_complex_identifier(
+        self, token: sqlparse.tokens.Token, index: int
+    ) -> bool:
+        """
+        Checks if token is a part of complex identifier like
+        <schema>.<table>.<column> or <table/sub_query>.<column>
+        """
+        return str(token) == "." or (
+            index + 1 < self.tokens_length
+            and str(self.non_empty_tokens[index + 1]) == "."
+        )
+
+    def _combine_qualified_names(self, index: int, token: SQLToken) -> None:
+        """
+        Combines names like <schema>.<table>.<column> or <table/sub_query>.<column>
+        """
+        value = token.value
+        is_complex = True
+        while is_complex:
+            value, is_complex = self._combine_tokens(index=index, value=value)
+            index = index - 2
+        token.value = value
+
+    def _combine_tokens(self, index: int, value: str) -> Tuple[str, bool]:
+        """
+        Checks if complex identifier is longer and follows back until it's finished
+        """
+        if index > 1 and str(self.non_empty_tokens[index - 1]) == ".":
+            prev_value = self.non_empty_tokens[index - 2].value.strip("`").strip('"')
+            value = f"{prev_value}.{value}"
+            return value, True
+        return value, False
+
+    def _get_sqlparse_tokens(self, parsed) -> None:
+        """
+        Flattens the tokens and removes whitespace
+        """
+        self.sqlparse_tokens = parsed[0].tokens
+        sqlparse_tokens = self._flatten_sqlparse()
+        self.non_empty_tokens = [
+            token
+            for token in sqlparse_tokens
+            if token.ttype is not Whitespace and token.ttype.parent is not Whitespace
+        ]
+        self.tokens_length = len(self.non_empty_tokens)
+
     def _flatten_sqlparse(self):
         for token in self.sqlparse_tokens:
             # sqlparse returns mysql digit starting identifiers as group
